import json
import numbers
import os
import sys
from pathlib import Path
import contextlib

from .file_helpers import read_contents, write_contents
from .mocodo_error import MocodoError


def safe_print_for_PHP(s):
    """ It seems that when called from PHP, Python is unable to guess correctly
        the encoding of the standard output. """
    try:
        print(s, file=sys.stdout)
    except UnicodeEncodeError:
        print(s.encode("utf8"), file=sys.stdout)


class Common:

    def __init__(self, params):
        self.params = params

    def output_success_message(self, path):
        return _('Output file "{filename}" successfully generated.').format(filename=path)

    def update_source(self, source):
        path = Path(f"{self.params['output_name']}.mcd")
        path.write_text(source)
        return _('Source file "{filename}" successfully updated.').format(filename=path)

    def load_input_file(self):
        for encoding in self.params["encodings"]:
            with contextlib.suppress(UnicodeError):
                self.encoding = encoding
                return read_contents(self.params["input"], encoding=encoding).replace('"', '')
        raise MocodoError(5, _('Unable to read "{filename}" with any of the following encodings: "{encodings}".').format(filename=self.params["input"], encodings= ", ".join(self.params["encodings"]))) # fmt: skip

    def update_input_file(self, source):
        if not source.startswith("%%mocodo"):
            source = f"%%mocodo\n{source}"
        write_contents(self.params["input"], source, encoding=self.encoding)
        safe_print_for_PHP(self.output_success_message(self.params["input"]))

    def load_style(self):
        
        def load_by_name(name):
            path = self.params[name] + ("" if self.params[name].endswith(".json") else ".json") 
            if os.path.exists(path):
                try:
                    return json.loads(read_contents(path))
                except:
                    raise MocodoError(3, _('Problem with "{name}" file "{path}".').format(name=name, path=path)) # fmt: skip
            path = os.path.join(self.params["script_directory"], "resources", name, path)
            try:
                return json.loads(read_contents(path))
            except:
                raise MocodoError(3, _('Problem with "{name}" file "{path}".').format(name=name, path=path)) # fmt: skip
        
        def may_apply_scaling(shapes):
            if self.params["scale"] == 1:
                return
            scale = self.params["scale"]
            for key in shapes:
                if key.endswith("font"):
                    shapes[key]["size"] = shapes[key]["size"] * scale
                elif not key.endswith("ratio") and isinstance(shapes[key], numbers.Number):
                    shapes[key] *= scale
        
        def ensure_margins_are_integer(shapes):
            # Some nasty failures are known to occur otherwise.
            for key in shapes:
                if "margin" in key:
                    shapes[key] = int(shapes[key] + 0.5)
            
        style = {}
        style.update(load_by_name("colors"))
        shapes = load_by_name("shapes")
        may_apply_scaling(shapes)
        ensure_margins_are_integer(shapes)
        style.update(shapes)
        style["transparent_color"] = None
        return style

    def dump_file(self, path, content):
        write_contents(path, content)
        safe_print_for_PHP(self.output_success_message(path))



from pathlib import Path

def read_contents(path, encoding="utf8"):
    return Path(path).read_text(encoding=encoding)

def write_contents(path, contents, encoding="utf8"):
    Path(path).write_text(contents, encoding=encoding)



import argparse
import contextlib
import importlib
import json
import os
import re
import shlex
import warnings
from base64 import b64encode
from itertools import takewhile
from pathlib import Path
from subprocess import PIPE, Popen

import pkg_resources
from IPython import get_ipython
from IPython.core.magic import Magics, line_cell_magic, magics_class
from IPython.display import HTML, SVG, Code, Image, Markdown, display

try:
    MOCODO_VERSION = pkg_resources.get_distribution("mocodo").version
except pkg_resources.DistributionNotFound:
    MOCODO_VERSION = "(unknown version)"  # For tests during CI

# ANSI color codes
OK = "\033[92m"
WARNING = "\033[1m\033[38;5;166m"
FAIL = "\033[1m\033[91m"
RESET = "\033[0m"

IPYTHON = get_ipython()

def update_cell(content):
    IPYTHON.set_next_input(content, replace=True)

PARAM_TEMPLATE = '''\
# You may edit and run the following lines
import json, pathlib
params = """\\
{stdoutdata}"""
try:
    json.loads(params)
except:
    raise RuntimeError("Invalid JSON. Check your syntax on https://jsonlint.com.")
pathlib.Path("{output_dir}/params.json").write_text(params);'''

OUTPUT_PART_HEADER = re.sub("  +", "", """
<div style="position: relative; height: 3ex; background-color: transparent">
  <hr style="margin: 1ex 0 0 0; border-top: 1px solid #BBB">
  <span style="position: absolute; right: 0; color: #BBB">
    <tt>
        {label}
    </tt>
  </span>
</div>
""")

def read_and_cleanup_text(path):
    text = path.read_text()
    text = re.sub(r"(?m).+Generated by Mocodo.+\n+", "", text)
    return text

def display_converted_file(path, hide_header):
    if not (hide_header and path.name.endswith("_mld.md")):
        display(Markdown(OUTPUT_PART_HEADER.format(label=path.relative_to(Path.cwd()))))
    extension=path.suffix[1:]
    if extension == "svg":
        # Fix a maximum width for SVG images:
        # https://stackoverflow.com/questions/51452569/how-to-resize-rescale-a-svg-graphic-in-an-ipython-jupyter-notebook
        svg = b64encode(path.read_bytes()).decode("utf8")
        # Use Markdown instead of HTML to avoid a grey background.
        display(Markdown(f'<img max-width="100%" src="data:image/svg+xml;base64,{svg}">'))
    elif extension == "md":
        display(Markdown(read_and_cleanup_text(path)))
    elif extension == "png":
        display(Image(filename=path, unconfined=False))
    elif extension == "html":
        text = path.read_text()
        text = re.sub('<!-- TO_BE_DELETED_BY_MOCODO_MAGIC -->.+\n', "", text)
        display(HTML(text))
    elif extension == "tex":
        display(Code(filename=path, language="latex"))
    elif extension == "url":
        print(path.read_text()) # make the link clickable
    elif extension == "tsv":
        try:
            df = importlib.import_module("pandas").read_csv(path, sep="\t", header=0, index_col=False)
            df = df.style.set_properties(**{'text-align': 'left'}).set_table_styles([ dict(selector='th', props=[('text-align', 'left')] ) ])
            display(df)
        except ImportError:
            display(Markdown(f"```\n{path.read_text()}\n```"))
    else:
        display(Markdown(f"```{extension}\n{read_and_cleanup_text(path)}\n```"))

@magics_class
class MocodoMagics(Magics):
    @staticmethod
    @line_cell_magic
    def mocodo(line, cell=""):
        """
        Mocodo IPython magic extension

        Magic methods:
            %mocodo [command line options]
            %%mocodo [command line options]
            < MCD ... >

        Usage:
            %load_ext mocodo
        """

        parser = argparse.ArgumentParser(add_help=False)
        parser.add_argument("--input", "-i")
        parser.add_argument("--output_dir")
        (args, remaining_args) = parser.parse_known_args(shlex.split(line))
        remaining_args = list(takewhile(lambda x: not x.startswith("#"), remaining_args))
        new_args = remaining_args[:]

        mocodo_notebook_dir = Path.cwd()
        if mocodo_notebook_dir.name != "mocodo_notebook":
            mocodo_notebook_dir = mocodo_notebook_dir / "mocodo_notebook"
            mocodo_notebook_dir.mkdir(parents=True, exist_ok=True)
        if not args.input:
            input_path = mocodo_notebook_dir / "sandbox.mcd"
            input_path.write_text(cell, encoding="utf8")
        else:
            input_path = Path(args.input)
            if not input_path.suffix:
                input_path = input_path.with_suffix(".mcd")

        if not args.output_dir:
            output_dir = mocodo_notebook_dir
        else:
            output_dir = Path(args.output_dir)
            output_dir.mkdir(parents=True, exist_ok=True)
        output_path_radical = output_dir / input_path.stem

        remaining_args.extend([
            "--input", str(input_path),
            "--output_dir", str(output_dir),
            "--is_magic",
        ]) # may override user's provided options

        process = Popen(["mocodo"] + remaining_args, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        stdoutdata, stderrdata = process.communicate()
        try:
            stdoutdata = stdoutdata.decode("utf8")
            stderrdata = stderrdata.decode("utf8")
        except:
            pass
        status = process.wait()

        if "--help" in remaining_args:
            print(stdoutdata)
            return
        
        if "--print_params" in remaining_args:
            update_cell(PARAM_TEMPLATE.format(stdoutdata=stdoutdata, output_dir=output_dir.relative_to(Path.cwd())))
            return
        
        response_path = Path(f"{output_path_radical}_response_for_magic_command.json")
        try: 
            response = json.loads(response_path.read_text())
        except (json.decoder.JSONDecodeError, FileNotFoundError):
            response = {}
        finally:
            with contextlib.suppress(FileNotFoundError): # From Python 3.8, use the missing_ok argument
                response_path.unlink()
        
        rewritten_source = response.get("rewritten_source", "")
        redirect_output = response.get("redirect_output", False)
        converted_file_paths = response.get("converted_file_paths", [])

        if status != 0 or stderrdata:
            message = stderrdata
            if rewritten_source:
                message = f"{rewritten_source}\n\n{message}"
            warnings.formatwarning = lambda x, *args, **kargs : str(x)
            warnings.warn(message)
            return
        
        if rewritten_source and not rewritten_source.startswith("%%mocodo"):
            new_args = " ".join(filter(lambda x: x not in response["args_to_delete"], new_args))
            new_args = response["opt_to_restore"] + new_args
            rewritten_source = f"%%mocodo{new_args}\n{rewritten_source}"
        
        for select in response.get("select", []):
            if select == "mcd":
                svg_path = output_path_radical.with_suffix(".svg")
                if svg_path.is_file() and input_path.stat().st_mtime <= svg_path.stat().st_mtime:
                    display(SVG(filename=svg_path))
            elif select == "rw":
                path = output_path_radical.with_suffix(".mcd")
                display(Markdown(OUTPUT_PART_HEADER.format(label=path.relative_to(output_dir))))
                print(rewritten_source or cell)
            elif select == "cv":
                for converted_file_path in converted_file_paths:
                    display_converted_file(Path(converted_file_path), hide_header=response.get("mld"))

        if redirect_output:
            if rewritten_source:
                update_cell(rewritten_source)
            if converted_file_paths:
                # Copy the last converted file source to the clipboard
                with contextlib.suppress(ImportError):
                    pyperclip = importlib.import_module("pyperclip")
                    converted_file_path = Path(converted_file_paths[-1])
                    pyperclip.copy(converted_file_path.read_text())
                    print(f'⧉ {OK}The contents of "{converted_file_path.relative_to(output_dir)}" has been copied to the clipboard.{RESET}')




try:
    IPYTHON.register_magics(MocodoMagics)
except AttributeError:
    pass # necessary for launching the tests
else:
    print(f"Mocodo {MOCODO_VERSION} loaded.")




from ..tools.parser_tools import parse_source

def run(source, id_prefix):
    if id_prefix is None:
        id_prefix = "id. "
    tree = parse_source(source)
    entity_name_refs =set(node.children[0].children[0].value for node in tree.find_data("entity_name_ref"))
    entity_name_defs = set(node.children[0].children[0].value for node in tree.find_data("entity_name_def"))
    new_entity_names = entity_name_refs - entity_name_defs
    new_entity_clauses = [source]
    for ent in new_entity_names:
        if ent.lower() != "date":
            clause = f"{ent}: {id_prefix}{ent.lower()}, "
        else:
            clause = f"{ent}: date"
        new_entity_clauses.append(clause)
    return "\n".join(new_entity_clauses)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mocodo', 'mocodo.convert', 'mocodo.dev', 'mocodo.rewrite', 'mocodo.tools']

package_data = \
{'': ['*'],
 'mocodo': ['resources/*',
            'resources/colors/*',
            'resources/i18n/*',
            'resources/lorem/*',
            'resources/relation_templates/*',
            'resources/shapes/*']}

install_requires = \
['requests>=2.22.0,<3.0.0']

entry_points = \
{'console_scripts': ['mocodo = mocodo.__main__:main']}

setup_kwargs = {
    'name': 'mocodo',
    'version': '4.0.13',
    'description': 'Modélisation Conceptuelle de Données. Nickel. Ni souris.',
    'long_description': '**Octobre 2023.** Mocodo [4](https://github.com/laowantong/mocodo/releases/tag/4.0.4) introduit la gestion manuelle et automatique des contraintes d\'optionalité et d\'unicité, améliore les interfaces graphique et en ligne de commande, et ajoute un grand nombre de fonctionnalités\xa0: coloration syntaxique, liens de partage, exportation en UML, en notation de Chen et _crow\'s foot_, génération de MCD aléatoires, décomposition des associations et autres opérations de réécriture.\n\n**Septembre 2022.** Mocodo 3 introduit l\'[héritage](https://rawgit.com/laowantong/mocodo/master/doc/fr_refman.html#Héritage-(ou-spécialisation)), l\'[agrégation](https://rawgit.com/laowantong/mocodo/master/doc/fr_refman.html#Agrégation-(ou-pseudo-entité)), les [calques](https://rawgit.com/laowantong/mocodo/master/doc/fr_refman.html#Héritage-(ou-spécialisation)), les [sorties PDF et PNG](https://rawgit.com/laowantong/mocodo/master/doc/fr_refman.html#Héritage-(ou-spécialisation)), etc.\xa0: [3.0](https://github.com/laowantong/mocodo/releases/tag/3.0), [3.1](https://github.com/laowantong/mocodo/releases/tag/3.1.0), [3.2](https://github.com/laowantong/mocodo/releases/tag/3.2.0).\n\n------\n\n[Documentation](https://rawgit.com/laowantong/mocodo/master/doc/fr_refman.html) sur une seule page, incluant la sortie de [`mocodo --help`](https://rawgit.com/laowantong/mocodo/master/doc/fr_refman.html#mocodo---help) et l\'[aide-mémoire des transformations](https://rawgit.com/laowantong/mocodo/master/doc/fr_refman.html#Aide-mémoire-des-arguments-de---transform).\n\n----\n\n![](logos/banner.svg)\n\nMocodo est un logiciel d\'aide à l\'enseignement et à l\'apprentissage des [bases de données relationnelles](https://fr.wikipedia.org/wiki/Base_de_données_relationnelle).\n\n- En entrée, il prend un [MCD](https://fr.wikipedia.org/wiki/Modèle_entité-association) (modèle conceptuel de données) décrit dans un langage dédié minimaliste.\n- En sortie, il produit un diagramme entité-association et, à la demande, un [MLD](https://fr.wikipedia.org/wiki/Merise_(informatique)#MLD_:_modèle_logique_des_données) (schéma relationnel, sous forme graphique ou textuelle), un [DDL](https://fr.wikipedia.org/wiki/Langage_de_définition_de_données) (requêtes SQL de création de la base), un [diagramme de classes UML](https://fr.wikipedia.org/wiki/Diagramme_de_classes), etc.\n- En bonus, il est capable de réarranger automatiquement votre MCD de façon esthétique, et de lui appliquer des opérations de réécriture qui vont du mondain (typographie) à l\'académique (décomposition d\'associations), en passant par le merveilleux (inférence de types, génération d\'exercices et d\'exemples).\n\nVous pouvez utiliser Mocodo\xa0:\n\n- à distance, sans rien installer, avec [Mocodo _online_](https://www.mocodo.net)\xa0;\n- en local, comme n\'importe quel programme Python\xa0;\n- dans un document [Jupyter Notebook](https://jupyter.org).\n\n## Tracé du modèle conceptuel\n\nCi-dessous, un exemple d\'utilisation sous Jupyter Notebook. L\'appel du programme est en première ligne\xa0; le texte-source proprement dit, lignes suivantes. En sortie, le diagramme conceptuel, égayé au passage par l\'option `--colors`\xa0:\n\n![png](doc/readme/ccp_mcd.png)\n\n![svg](doc/readme/ccp_mcd.svg)\n\n## Opérations de conversion\n\nOn peut récupérer ce texte-source avec `--input` pour lui appliquer diverses opérations. Ainsi, l\'appel suivant génère et affiche son MLD, son diagramme relationnel et son DDL\xa0:\n\n```\n%mocodo --input ccp --transform mld diagram ddl --colors desert\n```\n\n---\n\n- **CLIENT** (<ins>Réf. client</ins>, Nom, Prénom, Adresse)\n- **COMMANDE** (<ins>Num. commande</ins>, Date, Montant, _#Réf. client_)\n- **INCLURE** (<ins>_#Num. commande_</ins>, <ins>_#Réf. produit_</ins>, Quantité)\n- **PRODUIT** (<ins>Réf. produit</ins>, Libellé, Prix unitaire)\n\n---\n\n![svg](doc/readme/ccp_mld.svg)\n\n---\n\n```sql\nCREATE TABLE CLIENT (\n  PRIMARY KEY (ref_client),\n  ref_client VARCHAR(8) NOT NULL,\n  nom VARCHAR(255),\n  prenom VARCHAR(255),\n  adresse VARCHAR(255)\n);\n\nCREATE TABLE COMMANDE (\n  PRIMARY KEY (num_commande),\n  num_commande VARCHAR(8) NOT NULL,\n  date DATE,\n  montant DECIMAL(10,2),\n  ref_client VARCHAR(8) NOT NULL\n);\n\nCREATE TABLE INCLURE (\n  PRIMARY KEY (num_commande, ref_produit),\n  num_commande VARCHAR(8) NOT NULL,\n  ref_produit VARCHAR(8) NOT NULL,\n  quantite INTEGER\n);\n\nCREATE TABLE PRODUIT (\n  PRIMARY KEY (ref_produit),\n  ref_produit VARCHAR(8) NOT NULL,\n  libelle VARCHAR(50),\n  prix_unitaire DECIMAL(10,2)\n);\n\nALTER TABLE COMMANDE ADD FOREIGN KEY (ref_client) REFERENCES CLIENT (ref_client);\n\nALTER TABLE INCLURE ADD FOREIGN KEY (ref_produit) REFERENCES PRODUIT (ref_produit);\nALTER TABLE INCLURE ADD FOREIGN KEY (num_commande) REFERENCES COMMANDE (num_commande);\n```\n\nDans la suite, pour épargner la frappe, les options `--input` et `--transform` seront respectivement abrégées en `-i` et `-t`.\n\nLes opérations de conversion ne se limitent pas forcément au schéma relationnel. En voici une qui extrait un dictionnaire des données, par défaut sous la forme d\'un tableau Markdown à trois colonnes :\n\n```\n%mocodo -i ccp -t data_dict\n```\n\n---\n\n| Entité ou association | Libellé de l\'attribut | Type          |\n|:----------------------|:----------------------|:--------------|\n| CLIENT                | Adresse               | VARCHAR(255)  |\n| "                     | Nom                   | VARCHAR(255)  |\n| "                     | Prénom                | VARCHAR(255)  |\n| "                     | Réf. client           | VARCHAR(8)    |\n| COMMANDE              | Date                  | DATE          |\n| "                     | Montant               | DECIMAL(10,2) |\n| "                     | Num. commande         | VARCHAR(8)    |\n| INCLURE               | Quantité              | INTEGER       |\n| PRODUIT               | Libellé               | VARCHAR(50)   |\n| "                     | Prix unitaire         | DECIMAL(10,2) |\n| "                     | Réf. produit          | VARCHAR(8)    |\n\nUne autre qui transcrit le MCD dans la notation _crow\'s foot_ (`crow`) pour [Mermaid](http://mermaid.js.org) (`mmd`)\xa0:\n\n```\n%mocodo -i ccp -t crow:mmd\n```\n\n---\n\n```mmd\nerDiagram\n  CLIENT {\n    VARCHAR(8) ref_client PK\n    VARCHAR(255) nom\n    VARCHAR(255) prenom\n    VARCHAR(255) adresse\n  }\n  COMMANDE {\n    VARCHAR(8) num_commande PK\n    DATE date\n    DECIMAL(10-2) montant\n  }\n  INCLURE {\n    INTEGER quantite PK\n  }\n  PRODUIT {\n    VARCHAR(8) ref_produit PK\n    VARCHAR(50) libelle\n    DECIMAL(10-2) prix_unitaire\n  }\n  CLIENT ||--o{ COMMANDE: PASSER\n  INCLURE }|..|| COMMANDE: DF\n  INCLURE }o..|| PRODUIT: DF\n\n```\n\nLe rendu des diagrammes décrits dans des langages-tiers (comme Mermaid) n\'est pas directement pris en charge, mais peut être délégué (`--defer`) de façon transparente au service web approprié. Dans ce cas, c\'est la sortie graphique qui est affichée\xa0:\n\n```\n%mocodo -i ccp -t crow:mmd --defer\n```\n\n---\n\n![svg](doc/readme/ccp_erd_crow.svg)\n\n## Opérations de réécriture\n\nUne **réécriture** transforme un MCD Mocodo en un autre MCD Mocodo (au contraire d\'une **conversion**, qui produit un animal d\'une espèce différente).\n\nHeureusement, l\'utilisateur n\'a pas à réfléchir si la transformation qu\'il souhaite appliquer est une réécriture ou une conversion\xa0: dans les deux cas, il invoque `-t` (c\'est-à-dire `--transform`), et Mocodo se débrouille.\n\nEn guise de premier exemple de réécriture, mettons les noms des entités et associations (`boxes`) en majuscules, et les libellés (`labels`) en ASCII et _snake case_\xa0:\n\n```\n%mocodo -i ccp -t upper:boxes ascii:labels snake:labels --colors brewer+3\n```\n\n![svg](doc/readme/ccp_mcd_ascii.svg)\n\n---\n\n    %%mocodo --colors brewer+3\n    CLIENT: ref_client [VARCHAR(8)], nom [VARCHAR(255)], prenom [VARCHAR(255)], adresse [VARCHAR(255)]\n    PASSER, 0N CLIENT, 11 COMMANDE\n    COMMANDE: num_commande [VARCHAR(8)], date [DATE], montant [DECIMAL(10,2)]\n    INCLURE, 1N COMMANDE, 0N PRODUIT: quantite [INTEGER]\n    PRODUIT: ref_produit [VARCHAR(8)], libelle [VARCHAR(50)], prix_unitaire [DECIMAL(10,2)]\n\nRemarquez que l\'exécution d\'une réécriture affiche, au-dessous du diagramme, le code-source résultant. Celui-ci est précédé de la commande magique originale, _privée de l\'option `-i` et de toute opération de réécriture_. Ces dispositions permettent de continuer à travailler directement dessus si on le copie-colle dans une autre cellule.\n\nPlusieurs opérations de réécriture de nature sémantique sont également offertes. Par exemple, on peut décomposer un MCD quelconque en un MCD équivalent, mais n\'employant que des dépendances fonctionnelles et des entités faibles\xa0:\n\n```\n%mocodo -i ccp --select mcd -t explode:weak,arity=2 arrange:wide --seed=3 --colors brewer+3\n```\n\n![svg](doc/readme/ccp_mcd_explode.svg)\n\nNotez la sous-option `arrange:wide`. Elle a procédé à une réorganisation aléatoire des boîtes, ce que l\'insertion de deux nouvelles associations de dépendance fonctionnelles avait rendu nécessaire. Quant à l\'option `--seed=3`, elle garantit que le résultat sera le même à chaque exécution.\n\nPour vous familiariser avec Mocodo, le plus simple est d\'utiliser [sa version en ligne](https://www.mocodo.net). Vous pourrez ensuite vous plonger dans la [documentation](https://rawgit.com/laowantong/mocodo/master/doc/fr_refman.html).\n',
    'author': 'Aristide Grange',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://www.mocodo.net/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)
